﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;

  using Stack;

  /// <summary>
  /// Tells whether that stack is empty.
  /// </summary>
  [StorableClass]
  public abstract class EmptyExpression<T> : StatelessExpression {
    protected EmptyExpression() { }
    [StorableConstructor]
    protected EmptyExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return false;
    }

    public void Eval(IPushStack<T> stack, IPushStack<bool> booleanStack) {
      booleanStack.Push(stack.IsEmpty);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Exec, "EXEC.EMPTY", StackTypes.Boolean)]
  public class ExecEmptyExpression : EmptyExpression<Expression> {
    public ExecEmptyExpression() { }
    [StorableConstructor]
    protected ExecEmptyExpression(bool deserializing) : base(deserializing) { }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.ExecStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Code, "CODE.EMPTY", StackTypes.Boolean)]
  public class CodeEmptyExpression : EmptyExpression<Expression> {
    public CodeEmptyExpression() { }
    [StorableConstructor]
    protected CodeEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CodeStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.Integer, "INTEGER.EMPTY", StackTypes.Boolean)]
  public class IntegerEmptyExpression : EmptyExpression<long> {
    public IntegerEmptyExpression() { }
    [StorableConstructor]
    protected IntegerEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.EMPTY", StackTypes.Boolean)]
  public class FloatEmptyExpression : EmptyExpression<double> {
    public FloatEmptyExpression() { }
    [StorableConstructor]
    protected FloatEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.EMPTY", StackTypes.Boolean)]
  public class BooleanEmptyExpression : EmptyExpression<bool> {
    public BooleanEmptyExpression() { }
    [StorableConstructor]
    protected BooleanEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.EMPTY", StackTypes.Boolean)]
  public class CharEmptyExpression : EmptyExpression<char> {
    public CharEmptyExpression() { }
    [StorableConstructor]
    protected CharEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CharStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.EMPTY", StackTypes.Boolean)]
  public class StringEmptyExpression : EmptyExpression<string> {
    public StringEmptyExpression() { }
    [StorableConstructor]
    protected StringEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].EMPTY", StackTypes.Boolean)]
  public class IntegerVectorEmptyExpression : EmptyExpression<IReadOnlyList<long>> {
    public IntegerVectorEmptyExpression() { }
    [StorableConstructor]
    protected IntegerVectorEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.FloatVector, "FLOAT[].EMPTY", StackTypes.Boolean)]
  public class FloatVectorEmptyExpression : EmptyExpression<IReadOnlyList<double>> {
    public FloatVectorEmptyExpression() { }
    [StorableConstructor]
    protected FloatVectorEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].EMPTY", StackTypes.Boolean)]
  public class BooleanVectorEmptyExpression : EmptyExpression<IReadOnlyList<bool>> {
    public BooleanVectorEmptyExpression() { }
    [StorableConstructor]
    protected BooleanVectorEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.StringVector, "STRING[].EMPTY", StackTypes.Boolean)]
  public class StringVectorEmptyExpression : EmptyExpression<IReadOnlyList<string>> {
    public StringVectorEmptyExpression() { }
    [StorableConstructor]
    protected StringVectorEmptyExpression(bool deserializing) : base(deserializing) { }
    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack, interpreter.BooleanStack);
    }
  }
}
