﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System.Collections.Generic;

  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;

  using Interpreter;
  using Stack;

  [StorableClass]
  public abstract class RotateExpression<T> : StatelessExpression {
    protected RotateExpression() { }
    [StorableConstructor]
    protected RotateExpression(bool deserializing) : base(deserializing) { }

    protected void Eval(IPushStack<T> stack) {
      stack.Swap(3);
    }
  }


  [PushExpression(StackTypes.Integer, "INTEGER.ROT")]
  [StorableClass]
  public class IntegerRotateExpression : RotateExpression<long> {
    public IntegerRotateExpression() { }
    [StorableConstructor]
    protected IntegerRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerStack);
    }
  }

  [PushExpression(StackTypes.Float, "FLOAT.ROT")]
  [StorableClass]
  public class FloatRotateExpression : RotateExpression<double> {
    public FloatRotateExpression() { }
    [StorableConstructor]
    protected FloatRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatStack);
    }
  }

  [PushExpression(StackTypes.Boolean, "BOOLEAN.ROT")]
  [StorableClass]
  public class BooleanRotateExpression : RotateExpression<bool> {
    public BooleanRotateExpression() { }
    [StorableConstructor]
    protected BooleanRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanStack);
    }
  }

  [PushExpression(StackTypes.Name, "NAME.ROT")]
  [StorableClass]
  public class NameRotateExpression : RotateExpression<string> {
    public NameRotateExpression() { }
    [StorableConstructor]
    protected NameRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.NameStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.NameStack);
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.ROT", execIn: 3)]
  [StorableClass]
  public class ExecRotateExpression : RotateExpression<Expression> {
    public ExecRotateExpression() { }
    [StorableConstructor]
    protected ExecRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.ExecStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.ExecStack);
    }
  }

  [PushExpression(StackTypes.Code, "CODE.ROT")]
  [StorableClass]
  public class CodeRotateExpression : RotateExpression<Expression> {
    public CodeRotateExpression() { }
    [StorableConstructor]
    protected CodeRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CodeStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CodeStack);
    }
  }

  [PushExpression(StackTypes.Char, "CHAR.ROT")]
  [StorableClass]
  public class CharRotateExpression : RotateExpression<char> {
    public CharRotateExpression() { }
    [StorableConstructor]
    protected CharRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.CharStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.CharStack);
    }
  }

  [PushExpression(StackTypes.String, "STRING.ROT")]
  [StorableClass]
  public class StringRotateExpression : RotateExpression<string> {
    public StringRotateExpression() { }
    [StorableConstructor]
    protected StringRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringStack);
    }
  }

  [PushExpression(StackTypes.IntegerVector, "INTEGER[].ROT")]
  [StorableClass]
  public class IntegerVectorRotateExpression : RotateExpression<IReadOnlyList<long>> {
    public IntegerVectorRotateExpression() { }
    [StorableConstructor]
    protected IntegerVectorRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.IntegerVectorStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.IntegerVectorStack);
    }
  }

  [PushExpression(StackTypes.FloatVector, "FLOAT[].ROT")]
  [StorableClass]
  public class FloatVectorRotateExpression : RotateExpression<IReadOnlyList<double>> {
    public FloatVectorRotateExpression() { }
    [StorableConstructor]
    protected FloatVectorRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.FloatVectorStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.FloatVectorStack);
    }
  }

  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].ROT")]
  [StorableClass]
  public class BooleanVectorRotateExpression : RotateExpression<IReadOnlyList<bool>> {
    public BooleanVectorRotateExpression() { }
    [StorableConstructor]
    protected BooleanVectorRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.BooleanVectorStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.BooleanVectorStack);
    }
  }

  [PushExpression(StackTypes.StringVector, "STRING[].ROT")]
  [StorableClass]
  public class StringVectorRotateExpression : RotateExpression<IReadOnlyList<string>> {
    public StringVectorRotateExpression() { }
    [StorableConstructor]
    protected StringVectorRotateExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return interpreter.StringVectorStack.Count < 3;
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter.StringVectorStack);
    }
  }
}