﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Problem.BenchmarkSuite {
  using Common;
  using Configuration;
  using Core;

  using HeuristicLab.BenchmarkSuite;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;

  using Instances;
  using Persistence.Default.CompositeSerializers.Storable;
  using Stack;

  [StorableClass]
  [Creatable(CreatableAttribute.Categories.GeneticProgrammingProblems, Priority = 180)]
  [Item("Push Problem", "")]
  public class PushBenchmarkSuiteProblem : PushProblem, IProblemInstanceConsumer<ProblemData> {
    public PushBenchmarkSuiteProblem() : base(new PushBenchmarkSuiteEvaluator()) {
      Parameters.Add(PushEvaluator.DataParameter);
      Parameters.Add(PushEvaluator.DataBoundsParameter);
    }

    [StorableConstructor]
    protected PushBenchmarkSuiteProblem(bool deserializing)
      : base(deserializing) {
    }

    protected PushBenchmarkSuiteProblem(PushBenchmarkSuiteProblem original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PushBenchmarkSuiteProblem(this, cloner);
    }

    public new PushBenchmarkSuiteEvaluator PushEvaluator { get { return (PushBenchmarkSuiteEvaluator)base.PushEvaluator; } }

    public void Load(ProblemData data) {
      PushEvaluator.LoadData(data);

      Name = "Push Problem: " + data.Name;
      Description = data.Description;
      BestKnownQuality = data.BestResult;
      config.MaxPointsInProgram = data.MaxSize;
      config.EvalPushLimit = data.EvalLimit;
      config.ErcOptions = data.ErcOptions;
      config.FloatStringFormat = data.FloatStringFormat;

      config.SetEnabledStacks((StackTypes)data.EnabledDataTypes);

      Encoding.Bounds[0, 0] = 0;
      Encoding.Bounds[0, 1] = config.EnabledExpressions.Count;

      InitProgramLength = data.MaxSize / 2;
    }

    protected override PushSolution CreatePushSolution(
      PushProgram program,
      double bestQuality,
      IRandom random,
      IReadOnlyPushConfiguration config,
      IPushEvaluator evaluator) {
      return new PushBenchmarkSuiteSolution(program, bestQuality, random, (IReadOnlyPushConfiguration)config.Clone(), (PushBenchmarkSuiteEvaluator)PushEvaluator.Clone());
    }
  }
}
